"""This module will generate a COCO JSON file from the ZOD dataset."""

import json
import os
from functools import partial
from pathlib import Path
from typing import List, Tuple

import typer
from tqdm.contrib.concurrent import process_map

import zod.constants as constants
from zod import ZodFrames
from zod.anno.object import AnnotatedObject
from zod.constants import ALL_CLASSES, Anonymization
from zod.data_classes.frame import ZodFrame
from zod.utils.utils import str_from_datetime

# Map classes to categories, starting from 1
CATEGORY_NAME_TO_ID = {cls: i + 1 for i, cls in enumerate(ALL_CLASSES)}
OPEN_DATASET_URL = (
    "https://www.ai.se/en/data-factory/datasets/data-factory-datasets/zenseact-open-dataset"
)


def _convert_frame(
    frame: ZodFrame, classes: List[str], anonymization: constants.Anonymization, use_png: bool
) -> Tuple[dict, List[dict]]:

    objs: List[AnnotatedObject] = frame.get_annotation(constants.AnnotationProject.OBJECT_DETECTION)
    camera_frame = frame.info.get_key_camera_frame(anonymization=anonymization)
    file_name = camera_frame.filepath

    if anonymization == constants.Anonymization.ORIGINAL:
        file_name = file_name.replace(
            Anonymization.BLUR.value, constants.Anonymization.ORIGINAL.value
        )
    if use_png:
        file_name = file_name.replace(".jpg", ".png")

    image_dict = {
        "id": int(frame.info.id),
        "license": 1,
        "file_name": file_name,
        "height": camera_frame.height,
        "width": camera_frame.width,
        "date_captured": str_from_datetime(frame.info.keyframe_time),
    }
    anno_dicts = [
        {
            "id": int(frame.info.id) * 1000
            + obj_idx,  # avoid collisions by assuming max 1k objects per frame
            "image_id": int(frame.info.id),
            "category_id": CATEGORY_NAME_TO_ID[obj.name],
            "bbox": [round(val, 2) for val in obj.box2d.xywh.tolist()],
            "area": round(obj.box2d.area, 2),
            "iscrowd": obj.should_ignore_object(require_3d=False, require_eval=False),
        }
        for obj_idx, obj in enumerate(objs)
        if obj.name in classes
    ]
    return image_dict, anno_dicts


def generate_coco_json(
    dataset: ZodFrames,
    split: str,
    classes: List[str],
    anonymization: constants.Anonymization,
    use_png: bool,
) -> dict:
    """Generate COCO JSON file from the ZOD dataset."""
    assert split in ["train", "val"], f"Unknown split: {split}"
    frame_infos = [dataset[frame_id] for frame_id in dataset.get_split(split)]
    _convert_frame_w_classes = partial(
        _convert_frame, classes=classes, anonymization=anonymization, use_png=use_png
    )
    results = process_map(
        _convert_frame_w_classes,
        frame_infos,
        desc=f"Converting {split} frames",
        chunksize=50 if dataset._version == "full" else 1,
    )
    image_dicts, all_annos = zip(*results)
    anno_dicts = [anno for annos in all_annos for anno in annos]  # flatten
    coco_json = {
        "images": image_dicts,
        "annotations": anno_dicts,
        "info": {
            "description": "Zenseact Open Dataset",
            "url": OPEN_DATASET_URL,
            "version": dataset._version,  # TODO: add dataset versioning
            "year": 2022,
            "contributor": "ZOD team",
            "date_created": "2022/12/15",
        },
        "licenses": [
            {
                "url": "https://creativecommons.org/licenses/by-sa/4.0/",
                "name": "Attribution-ShareAlike 4.0 International (CC BY-SA 4.0)",
                "id": 1,
            },
        ],
        "categories": [
            {"supercategory": "object", "id": category_id, "name": category_name}
            for category_name, category_id in CATEGORY_NAME_TO_ID.items()
            if category_name in classes
        ],
    }
    return coco_json


# Use typer instead of argparse
def convert_to_coco(
    dataset_root: Path = typer.Option(
        ...,
        exists=True,
        dir_okay=True,
        writable=False,
        readable=True,
        resolve_path=True,
        help="Path to the root of the ZOD dataset.",
    ),
    output_dir: Path = typer.Option(
        ...,
        exists=True,
        dir_okay=True,
        writable=True,
        readable=True,
        resolve_path=True,
        help="Path to the output directory.",
    ),
    version: str = typer.Option("full", help="Version of the dataset to use. One of: full, mini."),
    anonymization: constants.Anonymization = typer.Option(
        constants.Anonymization.BLUR, help="Anonymization mode to use."
    ),
    use_png: bool = typer.Option(False, help="Whether to use PNG images instead of JPG."),
    classes: List[str] = typer.Option(
        ["Vehicle", "Pedestrian", "VulnerableVehicle"], help="Classes to include in the dataset."
    ),
):
    # check that classes are valid
    for cls in classes:
        if cls not in ALL_CLASSES:
            typer.echo(f"ERROR: Invalid class: {cls}.")
            raise typer.Exit(1)
    typer.echo(
        "Converting ZOD to COCO format. "
        f"Version: {version}, anonymization: {anonymization}, classes: {classes}"
    )

    zod_frames = ZodFrames(dataset_root, version)

    base_name = f"zod_{version}_{anonymization}"
    if use_png:
        base_name += "_png"

    os.makedirs(output_dir, exist_ok=True)

    coco_json_train = generate_coco_json(
        zod_frames, split="train", classes=classes, anonymization=anonymization, use_png=use_png
    )
    with open(os.path.join(output_dir, f"{base_name}_train.json"), "w") as f:
        json.dump(coco_json_train, f)

    coco_json_val = generate_coco_json(
        zod_frames, split="val", classes=classes, anonymization=anonymization, use_png=use_png
    )
    with open(os.path.join(output_dir, f"{base_name}_val.json"), "w") as f:
        json.dump(coco_json_val, f)

    typer.echo("Successfully converted ZOD to COCO format. Output files:")
    typer.echo(f"    train:  {output_dir}/{base_name}_train.json")
    typer.echo(f"    val:    {output_dir}/{base_name}_val.json")


if __name__ == "__main__":
    typer.run(convert_to_coco)



import numpy as np
import typer
from rich import print
from rich.panel import Panel

from zod import ZodFrames, ZodSequences
from zod.data_classes import LidarData
from zod.utils.utils import zfill_id

app = typer.Typer(no_args_is_help=True)

IMPORT_ERROR = (
    "zod does not ship with open3d, which is required for interactive point cloud "
    "visualization.\nPlease install it manually: pip install open3d"
)


def _visualize(data: LidarData):
    try:
        import open3d as o3d
    except ImportError:
        print(Panel(IMPORT_ERROR, title="Warning! Missing Dependency", border_style="red"))
        exit(1)
    from matplotlib import cm

    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(data.points)
    # color the pointcloud according to the timestamp
    # timestamps = data.timestamps
    # color = (timestamps - timestamps.min()) / (timestamps.max() - timestamps.min())
    color = np.log(data.intensity)
    # use a nice colormap
    pcd.colors = o3d.utility.Vector3dVector(cm.get_cmap("jet")(color)[:, :3])
    # Draw coordinate frame
    o3d.visualization.draw_geometries(
        [pcd, o3d.geometry.TriangleMesh.create_coordinate_frame(size=1.0)]
    )


@app.command(no_args_is_help=True)
def frames(
    dataset_root: str = typer.Option(..., help="Path to the dataset root"),
    version: str = typer.Option(..., help="Version of the dataset"),
    frame_id: int = typer.Option(..., help="Frame id to visualize"),
    num_before: int = typer.Option(0, help="Number of frames before the given frame id"),
    num_after: int = typer.Option(0, help="Number of frames after the given frame id"),
):
    """Visualize the lidar data for a given frame id."""
    zod_frames = ZodFrames(dataset_root=dataset_root, version=version)
    if zfill_id(frame_id) not in zod_frames.get_all_ids():
        raise ValueError(f"Frame id must be one of {zod_frames.get_all_ids()}.")
    frame = zod_frames[frame_id]
    data = frame.get_aggregated_lidar(num_before=num_before, num_after=num_after)
    _visualize(data)


@app.command(no_args_is_help=True)
def sequences(
    dataset_root: str = typer.Option(..., help="Path to the dataset root"),
    version: str = typer.Option(..., help="Version of the dataset"),
    sequence_id: int = typer.Option(..., help="Frame id to visualize"),
    start: int = typer.Option(0, help="Index of the first frame to visualize"),
    end: int = typer.Option(-1, help="Index of the last frame to visualize (-1 means last)"),
    downsampling: int = typer.Option(10, help="Downsampling factor (random point dropping)"),
):
    """Visualize the lidar data for a given frame id."""
    zod_sequences = ZodSequences(dataset_root=dataset_root, version=version)
    if zfill_id(sequence_id) not in zod_sequences.get_all_ids():
        raise ValueError(f"Frame id must be one of {zod_sequences.get_all_ids()}.")
    frame = zod_sequences[sequence_id]
    data = frame.get_aggregated_lidar(start=start, end=end)
    if downsampling > 1:
        typer.echo(f"Will subsample the point-cloud with a factor {downsampling}")
        indexes = np.random.choice(
            data.points.shape[0], size=data.points.shape[0] // downsampling, replace=False
        )
        data.points = data.points[indexes]
        data.intensity = data.intensity[indexes]
        data.timestamps = data.timestamps[indexes]
        data.diode_idx = data.diode_idx[indexes]
    _visualize(data)


@app.command(no_args_is_help=True)
def path(
    path: str = typer.Argument(..., help="Path to the lidar data"),
):
    """Visualize a given lidar file (npy)."""
    _visualize(LidarData.from_npy(path))


if __name__ == "__main__":
    app()



from typing import Tuple

import cv2
import numpy as np

from zod.constants import Camera
from zod.data_classes.box import Box3D
from zod.data_classes.calibration import Calibration


def render_3d_box(
    image: np.ndarray,
    box3d: Box3D,
    calib: Calibration,
    color: Tuple[int, int, int] = (255, 0, 0),
    line_thickness=2,
) -> None:
    """Render a 3d box on the image."""

    assert isinstance(box3d.frame, Camera), "Only support rendering 3d boxes in camera frames."

    points = box3d.project_into_camera(calib).astype(np.int32)
    center_im = points[0]
    corners_im = points[1:]

    # draw the center point
    cv2.circle(image, center_im, 5, color, -1)

    # draw the 3d bounding box
    cv2.line(
        image,
        (corners_im[0][0], corners_im[0][1]),
        (corners_im[1][0], corners_im[1][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[1][0], corners_im[1][1]),
        (corners_im[2][0], corners_im[2][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[2][0], corners_im[2][1]),
        (corners_im[3][0], corners_im[3][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[3][0], corners_im[3][1]),
        (corners_im[0][0], corners_im[0][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[4][0], corners_im[4][1]),
        (corners_im[5][0], corners_im[5][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[5][0], corners_im[5][1]),
        (corners_im[6][0], corners_im[6][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[6][0], corners_im[6][1]),
        (corners_im[7][0], corners_im[7][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[7][0], corners_im[7][1]),
        (corners_im[4][0], corners_im[4][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[0][0], corners_im[0][1]),
        (corners_im[4][0], corners_im[4][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[1][0], corners_im[1][1]),
        (corners_im[5][0], corners_im[5][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[2][0], corners_im[2][1]),
        (corners_im[6][0], corners_im[6][1]),
        color,
        line_thickness,
    )
    cv2.line(
        image,
        (corners_im[3][0], corners_im[3][1]),
        (corners_im[7][0], corners_im[7][1]),
        color,
        line_thickness,
    )



from .polygon_utils import overlay_polygon_on_image


def overlay_lane_markings_on_image(polygons, image, fill_color=(100, 0, 0), alpha=1.0):
    for polygon in polygons:
        image = overlay_polygon_on_image(polygon, image, fill_color, alpha)
    return image



"""Plot gps information."""

import json
from typing import List, Optional, Sequence

import numpy as np
import plotly.express as px

from zod.data_classes.oxts import EgoMotion

MAPS_STYLE = "open-street-map"
SIZE_MAX = 7
OPACITY_LEVEL = 1
DEFAULT_COLOR = "red"
DEFAULT_SIZE = 1


class PlotlyAutoZoomer:
    """Class to find zoom level automatically based on long / lat diff as reference."""

    LONG_LAT_DIFFS = (0.00001, 0.0001, 0.0005, 0.001, 0.01, 0.05, 0.1, 0.2, 0.3, 0.5, 1, 2, 3)
    ZOOM_LEVELS = (24, 20, 18, 17, 14, 12, 11, 10, 9, 8, 7, 6, 4)

    @classmethod
    def calc_zoom_level(cls, lats: List[float], longs: List[float]) -> int:
        """Calculate zoom level based on lats and lons values.

        Args:
            lats : list of latitudes
            longs : list of longitudes

        Returns:
            zoom level for plotly

        """
        lats = np.array(lats)
        longs = np.array(longs)
        lat_diff = lats.max() - lats.min()
        long_diff = longs.max() - longs.min()
        max_diff = max(lat_diff, long_diff)
        return np.round(np.interp(max_diff, cls.LONG_LAT_DIFFS, cls.ZOOM_LEVELS))


def plot_gps(
    longs: Sequence[float],
    lats: Sequence[float],
    colours: Optional[List[str]] = None,
    sizes: Optional[List[int]] = None,
    show: bool = False,
):
    """Plot gps coordinates on map.

    Args:
        longs : gps points longitudes
        lats : gps points latitudes
        colours : colour of points on the map
        sizes : size of points on the map
        show : show map on browser if True, otherwise wait for further plots

    """
    assert len(longs) == len(lats)
    if not colours:
        colours = [DEFAULT_COLOR] * len(longs)
    if not sizes:
        sizes = [DEFAULT_SIZE] * len(longs)
    fig = px.scatter_mapbox(
        lat=lats,
        lon=longs,
        color=colours,
        size=sizes,
        zoom=PlotlyAutoZoomer.calc_zoom_level(lats, longs),
        size_max=SIZE_MAX,
        opacity=OPACITY_LEVEL,
    )
    fig.update_layout(mapbox_style=MAPS_STYLE)
    fig.update_layout(margin={"r": 0, "t": 0, "l": 0, "b": 0})
    if show:
        fig.show()


def plot_gps_track_from_dataset_sequence(ego_motion: EgoMotion):
    """Plot GPS track on the map from dataset sequence.

    Args:
        oxts_values : OxTS values

    """
    raise NotImplementedError("This function got broken, need to fix to work with EgoMotion")
    longs, lats = oxts_values.pos_lon, oxts_values.pos_lat
    colors = ["blue"] * len(longs)
    sizes = [1] * len(longs)
    plot_gps(longs, lats, colors, sizes, show=True)


def show_gps_for_all_frames(filename: str):
    """Show GPS points for all extracted frames in dataset.

    Args:
        filename: path to JSON file containing GPS coordinates for all frames in dataset

    """
    with open(filename) as opened:
        content = json.load(opened)
    lats, lons = [], []
    vehicles = set()
    for frame_id, points in content.items():
        vehicles.add(frame_id.split("_")[0])
        lons.append(points[0])
        lats.append(points[1])
    print(
        f"Total frames in dataset: {len(content)}, vehicles: {vehicles}",
    )
    plot_gps(lons, lats, ["blue"] * len(lons), [1] * len(lons), show=True)



from typing import List, Sequence, Tuple

import cv2
import numpy as np

from zod.anno.object import AnnotatedObject
from zod.constants import Camera
from zod.data_classes.box import Box2D, Box3D
from zod.utils.visualization import render_3d_box

FONT_TYPE = cv2.FONT_HERSHEY_COMPLEX


def apply_scale(values: Sequence[float], scale_factor: float) -> Sequence[float]:
    """Apply scale to values."""
    if scale_factor:
        return tuple(map(lambda x: scale_factor * x, values))
    return values


def calc_iou(box1_corners, box2_corners):
    """Calculate IoU between two boxes.

    Args:
        box1_corners (tuple[tuple[float]]) : left-top and right-bottom points of first box
        box2_corners (tuple[tuple[float]]) : left-top and right-bottom points of second box

    Returns:
        iou (float) : IoU metric

    """
    # select inner box corners
    inner_left_coord = max(box1_corners[0][0], box2_corners[0][0])
    inner_top_coord = max(box1_corners[0][1], box2_corners[0][1])
    inner_right_coord = min(box1_corners[1][0], box2_corners[1][0])
    inner_bottom_coord = min(box1_corners[1][1], box2_corners[1][1])

    # compute the area of intersection rectangle
    inter_area = abs(
        max((inner_right_coord - inner_left_coord, 0))
        * max((inner_bottom_coord - inner_top_coord), 0)
    )
    if inter_area == 0:
        return 0
    # compute the area of both the prediction and ground-truth
    # rectangles
    box1_area = abs(
        (box1_corners[0][0] - box1_corners[1][0]) * (box1_corners[0][1] - box1_corners[1][1])
    )
    box2_area = abs(
        (box2_corners[0][0] - box2_corners[1][0]) * (box2_corners[0][1] - box2_corners[1][1])
    )

    # compute the intersection over union by taking the intersection
    # area and dividing it by the sum of prediction + ground-truth
    # areas - the interesection area
    iou = inter_area / float(box1_area + box2_area - inter_area)
    return iou


def overlay_object_2d_box_on_image(
    image, box2d: Box2D, color=(0, 0, 100), scale_factor=None, line_thickness=2
):
    """Visualize 2D box of annotated object on the image."""
    left_up = apply_scale(box2d.corners[0].astype(int), scale_factor)
    right_bottom = apply_scale(box2d.corners[2].astype(int), scale_factor)
    cv2.rectangle(image, left_up, right_bottom, color=color, thickness=line_thickness)
    return image


def overlay_object_3d_box_on_image(
    image,
    box3d: Box3D,
    calib,
    color=(0, 0, 100),
    scale_factor=None,
    line_thickness=2,
    camera: Camera = Camera.FRONT,
):
    """Visualize 2D box of annotated object on the image."""
    box3d.convert_to(camera, calib)
    render_3d_box(image, box3d, calib, color, line_thickness)
    return image


def overlay_object_properties_on_image(
    image: np.ndarray,
    annotation: AnnotatedObject,
    color: Tuple[int, int, int] = (255, 0, 0),
    properties_list: List[str] = ["name"],
    scale_factor: float = 1,
    text_areas: list = [],
    text_scale: float = 0.9,
    object_id=-1,
) -> np.ndarray:
    """Visualize properties values for object."""

    properties = {
        "name": annotation.name,
        "object_type": annotation.object_type,
        "occlusion_level": annotation.occlusion_level,
        "object_id": object_id,
    }

    left_up = annotation.box2d.corners[0].astype(int)

    x_min, y_min = apply_scale(left_up, scale_factor)
    (width, height), _ = cv2.getTextSize("text", FONT_TYPE, text_scale, 2)
    # check text bounding box area
    text_area = [[x_min, y_min - height], [x_min, y_min]]
    text_lines = []
    for i, property_name in enumerate(properties_list):
        if property_name in properties:
            text = f"{property_name}:{properties[property_name]}"
        else:
            text = f"{property_name}:None"
        text_lines.append(text)
        (width, height), _ = cv2.getTextSize(text, FONT_TYPE, text_scale, 2)
        text_area[1][0] = max(text_area[1][0], x_min + width)
        text_area[1][1] = text_area[1][1] + height
    if text_areas and any(calc_iou(text_area, rec) for rec in text_areas):
        return image

    text_areas.append(text_area)

    # draw text
    for i, text in enumerate(text_lines):
        y = y_min + i * height
        if i == 0:
            (width, height), _ = cv2.getTextSize(text, FONT_TYPE, text_scale, 2)
            text_area = ((x_min, y - height), (x_min + width, y))
            cv2.rectangle(image, *text_area, color=(255, 255, 255), thickness=1)
        cv2.putText(image, text, (x_min, y), FONT_TYPE, text_scale, color, thickness=2)
    return image



import json
import os.path as osp
from itertools import repeat
from pathlib import Path
from typing import Dict, List, Set, Tuple, Union

from tqdm.contrib.concurrent import process_map

from zod.constants import SEQUENCES, TRAIN, TRAINVAL_FILES, VAL, VERSIONS
from zod.data_classes.info import Information
from zod.data_classes.sequence import ZodSequence
from zod.utils.utils import zfill_id


def _create_sequence(sequence: dict, dataset_root: str) -> Information:
    info = Information.from_dict(sequence)
    info.convert_paths_to_absolute(dataset_root)
    return info


class ZodSequences:
    """ZOD Sequences."""

    def __init__(self, dataset_root: Union[Path, str], version: str):
        self._dataset_root = dataset_root
        self._version = version
        assert version in VERSIONS, f"Unknown version: {version}, must be one of: {VERSIONS}"
        self._train_sequences, self._val_sequences = self._load_sequences()
        self._sequences: Dict[str, Information] = {
            **self._train_sequences,
            **self._val_sequences,
        }

    def __len__(self) -> int:
        return len(self._sequences)

    def __getitem__(self, sequence_id: Union[int, str, slice]) -> ZodSequence:
        """Get sequence by id, which is a 6-digit zero-padded number. Ex: '000001'."""
        if isinstance(sequence_id, slice):
            return [self.__getitem__(i) for i in range(*sequence_id.indices(len(self)))]

        sequence_id = zfill_id(sequence_id)
        return ZodSequence(self._sequences[sequence_id])

    def __iter__(self):
        for frame_id in self._sequences:
            yield self.__getitem__(frame_id)

    def _load_sequences(self) -> Tuple[Dict[str, Information], Dict[str, Information]]:
        """Load sequences for the given version."""
        filename = TRAINVAL_FILES[SEQUENCES][self._version]

        with open(osp.join(self._dataset_root, filename), "r") as f:
            all_ids = json.load(f)

        train_sequences = process_map(
            _create_sequence,
            all_ids[TRAIN],
            repeat(self._dataset_root),
            chunksize=1,
            desc="Loading train sequences",
        )
        val_sequences = process_map(
            _create_sequence,
            all_ids[VAL],
            repeat(self._dataset_root),
            chunksize=1,
            desc="Loading val sequences",
        )

        train_sequences = {s.id: s for s in train_sequences}
        val_sequences = {s.id: s for s in val_sequences}

        return train_sequences, val_sequences

    def get_split(self, split: str) -> List[str]:
        """Get split by name (e.g. train / val)."""
        if split == TRAIN:
            return list(self._train_sequences.keys())
        elif split == VAL:
            return list(self._val_sequences.keys())
        else:
            raise ValueError(f"Unknown split: {split}, should be {TRAIN} or {VAL}")

    def get_all_ids(self) -> Set[str]:
        """Get all sequence ids."""
        return set(self._sequences.keys())



"""Geometry utility functions."""
from typing import Union

import numpy as np


def project_3d_to_2d_kannala(
    points_3d: np.ndarray, camera_matrix: np.ndarray, distorion_coefs: np.ndarray
) -> np.ndarray:
    """Project 3d points to 2d using Kannala-Brandt model.

    Args:
        points_3d: 3d points in shape (N, 3). These are assumed to be in the
            camera frame to which they are beeing projected into.
        camera_matrix: Camera matrix in shape (3, 4).
        distorion_coefs: Distortion coefficients in shape (4,).

    Returns:
        2d points in shape (N, 2).
    """
    homogenous_points_3d = np.concatenate((points_3d, np.ones((points_3d.shape[0], 1))), axis=1)
    norm_data = np.linalg.norm(homogenous_points_3d[:, :2], axis=1)
    radial = np.arctan2(norm_data, homogenous_points_3d[:, 2])
    radial2 = radial**2
    radial4 = radial2**2
    radial6 = radial4 * radial2
    radial8 = radial4**2
    distortion_angle = radial * (
        1
        + distorion_coefs[0] * radial2
        + distorion_coefs[1] * radial4
        + distorion_coefs[2] * radial6
        + distorion_coefs[3] * radial8
    )
    u_dist = distortion_angle * homogenous_points_3d[:, 0] / norm_data
    v_dist = distortion_angle * homogenous_points_3d[:, 1] / norm_data
    pos_u = camera_matrix[0, 0] * u_dist + camera_matrix[0, 2]
    pos_v = camera_matrix[1, 1] * v_dist + camera_matrix[1, 2]

    return np.stack((pos_u, pos_v), axis=-1)


def unproject_2d_to_3d_kannala(
    points_2d: np.ndarray,
    camera_matrix: np.ndarray,
    undistorion_coefs: np.ndarray,
    depth: Union[np.ndarray, float],
) -> np.ndarray:
    """Unproject 2d points to 3d using Kannala-Brandt model.

    Args:
        points_2d: 2d points in shape (N, 2).
        camera_matrix: Camera matrix in shape (3, 4).
        undistorion_coefs: Unistortion coefficients in shape (4,).
        depth: Depth of the points in the camera frame.

    Returns:
        3d points in shape (N, 3).
    """
    N = points_2d.shape[0]
    assert np.isscalar(depth) or depth.shape == (N,)

    out = np.ones((N, 3))
    pixel_x, pixel_y = points_2d[:, 0], points_2d[:, 1]
    focal_x, focal_y = camera_matrix[0, 0], camera_matrix[1, 1]
    principal_x, principal_y = camera_matrix[0, 2], camera_matrix[1, 2]

    out[:, 0] = (pixel_x - principal_x) / focal_x
    out[:, 1] = (pixel_y - principal_y) / focal_y
    rho = np.linalg.norm(out[:, :2], axis=1)
    phi = np.arctan2(out[:, 1], out[:, 0])
    theta = rho * (
        1
        + undistorion_coefs[0] * rho**2
        + undistorion_coefs[1] * rho**4
        + undistorion_coefs[2] * rho**6
        + undistorion_coefs[3] * rho**8,
    )
    out[:, 0] = np.sin(theta) * np.cos(phi) * depth
    out[:, 1] = np.sin(theta) * np.sin(phi) * depth
    out[:, 2] = np.cos(theta) * depth

    return out


def transform_points(points: np.ndarray, transform: np.ndarray) -> np.ndarray:
    """Transform points from one frame to another.

    Args:
        points: Points in shape (N, 3).
        transform: Transform matrix in shape (4, 4).
    """
    points_homogenous = np.concatenate((points, np.ones((points.shape[0], 1))), axis=1)
    transformed_points = np.matmul(points_homogenous, transform.T)
    # Normalize the points
    transformed_points = transformed_points[:, :3] / transformed_points[:, 3:]
    # Remove the last dimension
    transformed_points = transformed_points[:, :3]
    return transformed_points


def get_points_in_camera_fov(fov: np.ndarray, camera_data: np.ndarray) -> np.ndarray:
    """Get points that are present in camera field of view.

    Args:
        fov: camera field of view
        camera_data: data to filter inside the camera field of view

    Returns:
        points only visible in the camera

    """
    horizontal_fov, vertical_fov = fov
    if horizontal_fov == 0:
        return camera_data
    angles = np.rad2deg(np.arctan2(camera_data[:, 0], camera_data[:, 2]))
    mask = np.logical_and(angles > -horizontal_fov / 2, angles < horizontal_fov / 2)
    return camera_data[mask.flatten()]



from dataclasses import dataclass

import numpy as np
from pyquaternion import Quaternion


@dataclass
class Pose:
    """A general class describing some pose."""

    transform: np.ndarray

    @property
    def translation(self) -> np.ndarray:
        """Return the translation (array)."""
        return self.transform[:3, 3]

    @property
    def rotation(self) -> Quaternion:
        """Return the rotation as a quaternion."""
        return Quaternion(matrix=self.rotation_matrix)

    @property
    def rotation_matrix(self) -> np.ndarray:
        """Return the rotation matrix."""
        return self.transform[:3, :3]

    @property
    def inverse(self) -> "Pose":
        """Return the inverse of the pose."""
        return Pose(np.linalg.inv(self.transform))

    @classmethod
    def from_translation_rotation(
        cls, translation: np.ndarray, rotation_matrix: np.ndarray
    ) -> "Pose":
        """Create a pose from a translation and a rotation."""
        transform = np.eye(4, 4)
        transform[:3, :3] = rotation_matrix
        transform[:3, 3] = translation
        return cls(transform)

    @classmethod
    def identity(cls) -> "Pose":
        """Create an identity pose."""
        return cls(np.eye(4, 4))



# nuScenes dev-kit.
# Code written by Holger Caesar & Oscar Beijbom, 2019.

import abc
from collections import defaultdict
from typing import List, Tuple, Union

import numpy as np


class EvalBox(abc.ABC):
    """Abstract base class for data classes used during detection evaluation. Can be a prediction or ground truth."""

    def __init__(
        self,
        sample_token: str = "",
        translation: Tuple[float, float, float] = (0, 0, 0),
        size: Tuple[float, float, float] = (0, 0, 0),
        rotation: Tuple[float, float, float, float] = (0, 0, 0, 0),
        velocity: Tuple[float, float] = (0, 0),
        ego_translation: Tuple[float, float, float] = (
            0,
            0,
            0,
        ),  # Translation to ego vehicle in meters.
        num_pts: int = -1,
    ):  # Nbr. LIDAR or RADAR inside the box. Only for gt boxes.

        # Assert data for shape and NaNs.
        assert type(sample_token) == str, "Error: sample_token must be a string!"

        assert len(translation) == 3, "Error: Translation must have 3 elements!"
        assert not np.any(np.isnan(translation)), "Error: Translation may not be NaN!"

        assert len(size) == 3, "Error: Size must have 3 elements!"
        assert not np.any(np.isnan(size)), "Error: Size may not be NaN!"

        assert len(rotation) == 4, "Error: Rotation must have 4 elements!"
        assert not np.any(np.isnan(rotation)), "Error: Rotation may not be NaN!"

        # Velocity can be NaN from our database for certain annotations.
        assert len(velocity) == 2, "Error: Velocity must have 2 elements!"

        assert len(ego_translation) == 3, "Error: Translation must have 3 elements!"
        assert not np.any(np.isnan(ego_translation)), "Error: Translation may not be NaN!"

        assert type(num_pts) == int, "Error: num_pts must be int!"
        assert not np.any(np.isnan(num_pts)), "Error: num_pts may not be NaN!"

        # Assign.
        self.sample_token = sample_token
        self.translation = translation
        self.size = size
        self.rotation = rotation
        self.velocity = velocity
        self.ego_translation = ego_translation
        self.num_pts = num_pts

    @property
    def ego_dist(self) -> float:
        """Compute the distance from this box to the ego vehicle in 2D."""
        return np.sqrt(np.sum(np.array(self.ego_translation[:2]) ** 2))

    def __repr__(self):
        return str(self.serialize())

    @abc.abstractmethod
    def serialize(self) -> dict:
        pass

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, content: dict):
        pass


EvalBoxType = Union["DetectionBox", "TrackingBox"]


class EvalBoxes:
    """Data class that groups EvalBox instances by sample."""

    def __init__(self):
        """
        Initializes the EvalBoxes for GT or predictions.
        """
        self.boxes = defaultdict(list)

    def __repr__(self):
        return "EvalBoxes with {} boxes across {} samples".format(
            len(self.all), len(self.sample_tokens)
        )

    def __getitem__(self, item) -> List[EvalBoxType]:
        return self.boxes[item]

    def __eq__(self, other):
        if not set(self.sample_tokens) == set(other.sample_tokens):
            return False
        for token in self.sample_tokens:
            if not len(self[token]) == len(other[token]):
                return False
            for box1, box2 in zip(self[token], other[token]):
                if box1 != box2:
                    return False
        return True

    def __len__(self):
        return len(self.boxes)

    @property
    def all(self) -> List[EvalBoxType]:
        """Returns all EvalBoxes in a list."""
        ab = []
        for sample_token in self.sample_tokens:
            ab.extend(self[sample_token])
        return ab

    @property
    def sample_tokens(self) -> List[str]:
        """Returns a list of all keys."""
        return list(self.boxes.keys())

    def add_boxes(self, sample_token: str, boxes: List[EvalBoxType]) -> None:
        """Adds a list of boxes."""
        self.boxes[sample_token].extend(boxes)

    def serialize(self) -> dict:
        """Serialize instance into json-friendly format."""
        return {key: [box.serialize() for box in boxes] for key, boxes in self.boxes.items()}

    @classmethod
    def deserialize(cls, content: dict, box_cls):
        """
        Initialize from serialized content.
        :param content: A dictionary with the serialized content of the box.
        :param box_cls: The class of the boxes, DetectionBox or TrackingBox.
        """
        eb = cls()
        for sample_token, boxes in content.items():
            eb.add_boxes(sample_token, [box_cls.deserialize(box) for box in boxes])
        return eb


class MetricData(abc.ABC):
    """Abstract base class for the *MetricData classes specific to each task."""

    @abc.abstractmethod
    def serialize(self):
        """Serialize instance into json-friendly format."""
        pass

    @classmethod
    @abc.abstractmethod
    def deserialize(cls, content: dict):
        """Initialize from serialized content."""
        pass



# (c) Copyright 2021-2022 Zenseact AB
"""Color labeler using some colormaps from OpenCV."""

from typing import Callable, Union

import cv2
import matplotlib as mpl

# pylint: disable=unused-import
import matplotlib.cm
import numpy as np


def create_matplotlib_colormap(src: np.ndarray, colormap_type: str) -> np.ndarray:
    """Create colormap image from matplotlib.

    Args:
        src : source image with specified shape
        colormap_type : colormap type

    Returns:
        array with colored values regarding specified colormap

    """
    distance_list = src[:, 0, 0]

    # use the coolwarm colormap that is built-in, and goes from blue to red
    cmap = getattr(mpl.cm, colormap_type)

    # convert your distances to color coordinates
    color_list = cmap(distance_list)
    return (np.array(color_list) * 255)[:, :3].astype("uint8").reshape(-1, 1, 3)


def create_opencv_colormap(src: np.ndarray, colormap_type: int) -> np.ndarray:
    """Create colormap image from matplotlib.

    Args:
        src : source image with specified shape
        colormap_type : colormap type

    Returns:
        array with colored values regarding specified colormap

    """
    return cv2.applyColorMap(src, colormap_type)


class ColorLabeler:
    """Class for labeling with color due to Colormap."""

    def __init__(
        self,
        map_type: Union[str, int] = cv2.COLORMAP_RAINBOW,
        map_creator: Callable = create_opencv_colormap,
        max_value: int = 256,
        normalized: bool = False,
    ):
        """Init.

        Args:
            map_type (int) : openCV colormap type
            max_value (int) : max value of color index
            normalized (boolean) : if True then all RGB values are in [0..1]

        """
        self.color_map_image_ = np.zeros((256, 30, 3), dtype="uint8")
        src = np.zeros((256, 1, 1), dtype="uint8")
        src[:, 0, 0] = range(0, 256)
        self.color_map_image_ = map_creator(src, map_type)
        self.mapsize = max_value
        if normalized:
            self.color_map_image_ = self.color_map_image_.astype("float32") / 255

    def label_to_color(self, label):
        """Convert label to color."""
        return tuple(
            int(val)
            for val in self.color_map_image_[int((label * 255) // self.mapsize) % 255, 0, :]
        )

    def label_to_color_norm(self, value):
        """Convert label to cv2 color."""
        return [int(val) for val in self.color_map_image_[int(value * 255), 0, :]]

    def label_to_color_id(self, color_idx):
        """Convert label to color."""
        return self.color_map_image_[color_idx, 0, :]

    def get_maxcolor(self):
        """Get maxcolor."""
        return self.mapsize

    def get_colormap(self):
        """Get colormap."""
        return self.color_map_image_

    def __call__(self, label):
        """Call for converting label to color."""
        return self.label_to_color(label)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zod',
 'zod.anno',
 'zod.anno.tsr',
 'zod.cli',
 'zod.data_classes',
 'zod.eval',
 'zod.eval.detection',
 'zod.eval.detection.nuscenes_eval.common',
 'zod.eval.detection.nuscenes_eval.detection',
 'zod.utils',
 'zod.visualization']

package_data = \
{'': ['*'], 'zod.eval.detection': ['nuscenes_eval/*']}

install_requires = \
['dataclass-wizard>=0.22.2',
 'h5py>=3.1',
 'numpy-quaternion>=2022.4.2',
 'numpy>=1.19,<2.0',
 'pillow>=7',
 'pyquaternion>=0.9',
 'scipy>=1.5,<2.0',
 'tqdm>=4.60']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata', 'typing-extensions'],
 'all': ['typer[all]>=0.7.0',
         'dropbox>=11.36.0',
         'opencv-python>=4',
         'matplotlib>=3',
         'plotly>=5,<6',
         'pandas>=1.3,<2.0',
         'notebook>=5',
         'imageio>=2,<3'],
 'cli': ['typer[all]>=0.7.0', 'dropbox>=11.36.0']}

entry_points = \
{'console_scripts': ['zod = zod.cli.main:app']}

setup_kwargs = {
    'name': 'zod',
    'version': '0.1.0',
    'description': 'Zenseact Open Dataset',
    'long_description': '# Zenseact Open Dataset\nThe Zenseact Open Dataset (ZOD) is a large multi-modal autonomous driving dataset developed by a team of researchers at [Zenseact](https://zenseact.com/). The dataset is split into three categories: *Frames*, *Sequences*, and *Drives*. For more information about the dataset, please refer to our [coming soon](), or visit our [website](https://zod.zenseact.com).\n\n## Examples\nFind examples of how to use the dataset in the [examples](examples/) folder. Here you will find a set of jupyter notebooks that demonstrate how to use the dataset, as well as an example of how to train an object detection model using [Detectron2](https://github.com/facebookresearch/detectron2).\n\n## Installation\n\nThe install the library with minimal dependencies, for instance to be used in a training environment without need for interactivity och visualization, run:\n```bash\npip install zod\n```\n\nTo install the library along with the CLI, which can be used to download the dataset, convert between formats, and perform visualization, run:\n```bash\npip install "zod[cli]"\n```\n\nTo install the full devkit, with the CLI and all dependencies, run:\n```bash\npip install "zod[all]"\n```\n\n## Download using the CLI\n\nThis is an example of how to download the ZOD Frames mini-dataset using the CLI. Prerequisites are that you have applied for access and received a download link. To download the mini-dataset, run:\n```bash\nzod download --url "<download-link>" --output-dir <path/to/outputdir> frames --mini\n```\nsimilarly, to download the full dataset (including all lidar scans before and after the keyframe), run:\n```bash\nzod download --url "<download-link>" --output-dir <path/to/outputdir> frames --lidar --num-scans-before -1 --num-scans-after -1 --oxts --images --blur --dnat --calibrations --annotations\n```\nthis will download all the previous and future lidar scans (as `num-scans-before=-1` and `num-scans-after=-1`), the OxTS data, the images (with both the blur and DNAT anonymization), the calibration files, the annotations, and all other necessary files. If you dont want any previous or future lidar scans, run:\n```bash\nzod download --url "<download-link>" --output-dir <path/to/outputdir> frames --lidar --num-scans-before 0 --num-scans-after 0 --oxts --images --blur --dnat --calibrations --annotations\n```\n\nFor a full list of options for ZOD download, run:\n```bash\nzod download --help\nzod download --url="<url>" --output-dir=<dir> frames --help\nzod download --url="<url>" --output-dir=<dir> sequences --help\n```\ndepending on which dataset you want to download.\n\n\n## Anonymization\nTo preserve privacy, the dataset is anonymized. The anonymization is performed by [brighterAI](https://brighter.ai/), and we provide two separate modes of anonymization: deep fakes (DNAT) and blur. In our paper, we show that the performance of an object detector is not affected by the anonymization method. For more details regarding this experiment, please refer to our [coming soon]().\n\n## Citation\nIf you publish work that uses Zenseact Open Dataset, please cite: [coming soon]()\n\n```\n@misc{zod2021,\n  author = {TODO},\n  title = {Zenseact Open Dataset},\n  year = {2023},\n  publisher = {TODO},\n  journal = {TODO},\n```\n\n## Contact\nFor questions about the dataset, please [Contact Us](mailto:opendataset@zenseact.com).\n\n## Contributing\nWe welcome contributions to the development kit. If you would like to contribute, please open a pull request.\n\n## License\n**Dataset**:\nThis dataset is the property of Zenseact AB (© 2023 Zenseact AB) and is licensed under [CC BY-SA 4.0](https://creativecommons.org/licenses/by-sa/4.0/). Any public use, distribution, or display of this dataset must contain this notice in full:\n\n> For this dataset, Zenseact AB has taken all reasonable measures to remove all personally identifiable information, including faces and license plates. To the extent that you like to request the removal of specific images from the dataset, please contact [privacy@zenseact.com](mailto:privacy@zenseact.com).\n\n\n**Development kit**:\nThis development kit is the property of Zenseact AB (© 2023 Zenseact AB) and is licensed under [MIT](https://opensource.org/licenses/MIT).\n',
    'author': 'Zenseact',
    'author_email': 'opendataset@zenseact.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://zod.zenseact.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



"""Relevant constants for ZOD."""

from enum import Enum
from typing import Union

# Python 3.7 compatibility
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

# Dataset paths
FRAMES = "single_frames"
SEQUENCES = "sequences"
DRIVES = "drives"
DB_DATE_STRING_FORMAT_W_MICROSECONDS = "%Y-%m-%dT%H:%M:%S.%f%z"

# Train-val splits
TRAIN = "train"
VAL = "val"
FULL = "full"
MINI = "mini"
TRAINVAL_FILES = {
    FRAMES: {
        FULL: f"trainval_frames_full.json",
        MINI: f"trainval_frames_mini.json",
    },
    SEQUENCES: {
        FULL: f"trainval_sequences_full.json",
        MINI: f"trainval_sequences_mini.json",
    },
}
SPLIT_FILES = {
    SEQUENCES: {
        FULL: {
            TRAIN: "splits/sequences_full_train.txt",
            VAL: "splits/sequences_full_val.txt",
        },
        MINI: {
            TRAIN: "splits/sequences_mini_train.txt",
            VAL: "splits/sequences_mini_val.txt",
        },
    },
    FRAMES: {
        FULL: {
            TRAIN: "splits/frames_full_train.txt",
            VAL: "splits/frames_full_val.txt",
        },
        MINI: {
            TRAIN: "splits/frames_mini_train.txt",
            VAL: "splits/frames_mini_val.txt",
        },
    },
}


VERSIONS = (FULL, MINI)


class Anonymization(Enum):
    BLUR = "blur"
    DNAT = "dnat"
    ORIGINAL = "original"


class AnnotationProject(Enum):
    OBJECT_DETECTION = "object_detection"
    LANE_MARKINGS = "lane_markings"
    TRAFFIC_SIGNS = "traffic_signs"
    EGO_ROAD = "ego_road"
    ROAD_CONDITION = "road_condition"


### Coordinate Frames ###

EGO = "ego"


class Camera(Enum):
    FRONT = "front"


class Lidar(Enum):
    VELODYNE = "velodyne"


CoordinateFrame = Union[Camera, Lidar, Literal[EGO]]


### Evaluation ###

EVALUATION_FRAME = Lidar.VELODYNE
ALL_CLASSES = [
    "Vehicle",
    "VulnerableVehicle",
    "Pedestrian",
    "Animal",
    "PoleObject",
    "Inconclusive",
    "TrafficBeacon",
    "TrafficSign",
    "TrafficSignal",
    "TrafficGuide",
    "DynamicBarrier",
]

EVALUATION_CLASSES = [
    "Vehicle",
    "VulnerableVehicle",
    "Pedestrian",
    "TrafficSign",
    "TrafficSignal",
]



"""Utility functions for Zod."""

import os
from datetime import datetime
from typing import Union

from zod.constants import DB_DATE_STRING_FORMAT_W_MICROSECONDS


def datetime_from_str(timestamp: str) -> datetime:
    """Convert a timestamp string to a datetime object."""
    return datetime.strptime(timestamp, DB_DATE_STRING_FORMAT_W_MICROSECONDS)


def str_from_datetime(timestamp: datetime) -> str:
    """Convert a datetime object to a timestamp string."""
    return timestamp.strftime(DB_DATE_STRING_FORMAT_W_MICROSECONDS)


def parse_datetime_from_filename(filename: str) -> datetime:
    """Parse a timestamp from a filename."""
    timestamp_str = os.path.splitext(os.path.basename(filename))[0].split("_")[-1]
    return datetime_from_str(timestamp_str)


def zfill_id(frame_id: Union[int, str]) -> str:
    if isinstance(frame_id, int):
        frame_id = str(frame_id)
    return frame_id.zfill(6)



"""Module to perform OxTS extraction and visualize GPS track projection on image plane."""
import cv2
import numpy as np

from zod.constants import Camera
from zod.data_classes.calibration import Calibration
from zod.data_classes.oxts import EgoMotion
from zod.utils.geometry import get_points_in_camera_fov, project_3d_to_2d_kannala, transform_points


def visualize_oxts_on_image(
    oxts: EgoMotion, key_timestamp, calibs: Calibration, image, camera=Camera.FRONT
):
    """Visualize oxts track on image plane."""

    # get pose at key frame
    current_pose = oxts.get_poses(key_timestamp)
    all_poses = oxts.poses

    # transform all poses to the current pose
    transformed_poses = np.linalg.pinv(current_pose) @ all_poses

    # get the positions from the transformed poses
    points = transformed_poses[:, :3, -1]

    # let's filter out the points that are behind the camera
    points = points[points[:, 0] > 0]

    # transform point to camera coordinate system
    T_inv = np.linalg.pinv(calibs.get_extrinsics(camera).transform)
    camerapoints = transform_points(points[:, :3], T_inv)
    print(f"Number of points: {points.shape[0]}")

    # filter points that are not in the camera field of view
    points_in_fov = get_points_in_camera_fov(calibs.cameras[camera].field_of_view, camerapoints)
    print(f"Number of points in fov: {len(points_in_fov)}")

    # project points to image plane
    xy_array = project_3d_to_2d_kannala(
        points_in_fov,
        calibs.cameras[camera].intrinsics[..., :3],
        calibs.cameras[camera].distortion,
    )

    for i in range(xy_array.shape[0]):
        cv2.circle(image, (int(xy_array[i, 0]), int(xy_array[i, 1])), 3, (255, 0, 0), -1)

    return image



from dataclasses import dataclass

from zod.data_classes.box import Box2D


@dataclass
class TrafficSignAnnotation:
    """Traffic sign dataclass."""

    bounding_box: Box2D
    traffic_sign_class: str
    traffic_sign_idx: int
    occlusion_ratio: str
    annotation_uuid: str
    electronic_sign: bool
    uuid: str



import json
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Tuple, Union

import h5py
import numpy as np
import quaternion
from scipy.interpolate import interp1d

OXTS_TIMESTAMP_OFFSET = datetime(1980, 1, 6, tzinfo=timezone.utc).timestamp()


@dataclass
class EgoMotion:
    poses: np.ndarray  # Nx4x4
    velocities: np.ndarray  # Nx3
    accelerations: np.ndarray  # Nx3
    angular_rates: np.ndarray  # Nx3
    timestamps: np.ndarray  # N - epoch time in seconds (float64)
    origin_lat_lon: Tuple[float, float]  # (lat, lon) of the origin (which pose are relative to)

    def __len__(self) -> int:
        return len(self.timestamps)

    def get_poses(self, target_ts: Union[np.ndarray, float]) -> np.ndarray:
        """Interpolate neighbouring poses to find pose for each target timestamp.
        Args:
            target_ts: [N] timestamps for which to find a pose by interpolation.
        Return:
            [N, 4, 4] array of interpolated poses for each timestamp.
        """
        selfmin, selfmax = np.min(self.timestamps), np.max(self.timestamps)
        assert (selfmin <= np.min(target_ts)) and (
            selfmax > np.max(target_ts)
        ), f"targets not between pose timestamps, must be [{selfmin}, {selfmax})"
        # TODO: check if the timestamp exists, then we don't need to interpolate
        # else interpolate according to below.
        closest_idxs = self.timestamps.searchsorted(
            target_ts, side="right", sorter=self.timestamps.argsort()
        )
        time_diffs = target_ts - self.timestamps[closest_idxs - 1]
        total_times = self.timestamps[closest_idxs] - self.timestamps[closest_idxs - 1]
        fractions = time_diffs / total_times
        return interpolate_transforms(
            self.poses[closest_idxs - 1], self.poses[closest_idxs], fractions
        )

    def interpolate(self, timestamps: np.ndarray) -> "EgoMotion":
        """Interpolate ego motion to find ego motion for each target timestamp.

        Args:
            timestamps: [N] timestamps for which to find ego motion by interpolation.
        Return:
            EgoMotion object with interpolated ego motion for each timestamp.
        """
        poses = self.get_poses(timestamps)
        velocities = interpolate_vectors(self.velocities, self.timestamps, timestamps)
        accelerations = interpolate_vectors(self.accelerations, self.timestamps, timestamps)
        angular_rates = interpolate_vectors(self.angular_rates, self.timestamps, timestamps)

        return EgoMotion(
            poses=poses,
            velocities=velocities,
            accelerations=accelerations,
            angular_rates=angular_rates,
            timestamps=timestamps,
            origin_lat_lon=self.origin_lat_lon,
        )

    @classmethod
    def from_oxts_path(cls, file_path: str) -> "EgoMotion":
        """Load ego motion from a sequence or frame oxts file."""
        with h5py.File(file_path, "r") as file:
            return cls(
                poses=file["poses"][()],
                accelerations=np.stack(
                    [file["accelerationX"], file["accelerationY"], file["accelerationZ"]], axis=1
                ),
                velocities=np.stack(
                    [file["velForward"][()], file["velLateral"][()], -file["velDown"][()]], axis=1
                ),
                angular_rates=np.stack(
                    [file["angularRateX"], file["angularRateY"], file["angularRateZ"]], axis=1
                ),
                timestamps=OXTS_TIMESTAMP_OFFSET + file["timestamp"][()] + file["leapSeconds"][()],
                origin_lat_lon=(file["posLat"][0], file["posLon"][0]),
            )

    @classmethod
    def from_json_path(cls, json_path: str) -> "EgoMotion":
        """Load ego motion from a json file."""
        with open(json_path, "r") as file:
            data = json.load(file)
        return cls(
            poses=np.array(data["poses"]),
            velocities=np.array(data["velocities"]),
            accelerations=np.array(data["accelerations"]),
            angular_rates=np.array(data["angular_rates"]),
            timestamps=np.array(data["timestamps"]),
            origin_lat_lon=tuple(data["origin_lat_lon"]),
        )

    def to_json(self) -> dict:
        """Save ego motion to a json file."""
        return {
            "poses": self.poses.tolist(),
            "velocities": self.velocities.tolist(),
            "accelerations": self.accelerations.tolist(),
            "angular_rates": self.angular_rates.tolist(),
            "timestamps": self.timestamps.tolist(),
            "origin_lat_lon": self.origin_lat_lon,
        }


def interpolate_transforms(transform_1: np.ndarray, transform_2: np.ndarray, fractions: np.ndarray):
    """Interpolate between two transforms.

    Args:
        transform_1: First transform (Mx4x4).
        transform_2: Second transform (Mx4x4).
        fraction: Fraction of interpolation (M).

    Returns:
        Interpolated transform(s) (Mx4x4).
    """
    R_a, t_a = transform_1[..., :3, :3], transform_1[..., :3, 3]
    R_b, t_b = transform_2[..., :3, :3], transform_2[..., :3, 3]
    # Alternative would be to convert to euler angles and interpolate, but this is more accurate.
    R_interpolated = quaternion.as_rotation_matrix(
        np.slerp_vectorized(
            quaternion.from_rotation_matrix(R_a),
            quaternion.from_rotation_matrix(R_b),
            fractions,
        )
    )
    t_interp = t_a + fractions[..., None] * (t_b - t_a)
    transform = np.concatenate(
        [
            np.concatenate([R_interpolated, t_interp[..., None]], axis=-1),
            np.broadcast_to(np.array([0, 0, 0, 1]), R_interpolated.shape[:-2] + (1, 4)),
        ],
        axis=-2,
    )
    return transform


def interpolate_vectors(
    values: np.ndarray, source_timestamps: np.ndarray, target_timestamps: np.ndarray
) -> np.ndarray:
    """Interpolate vectors to find vector for each (sorted) target timestamp.

    Args:
        values: [N, X] array of values.
        source_timestamps: [N] array of timestamps.
        target_timestamps: [M] array of timestamps.
    Returns:
        [M, X] array of interpolated values.
    """
    source_min_ts, source_max_ts = np.min(source_timestamps), np.max(source_timestamps)
    target_min_ts, target_max_ts = np.min(target_timestamps), np.max(target_timestamps)
    assert source_min_ts <= target_min_ts, "Target timestamps must be after source timestamps"
    assert source_max_ts >= target_max_ts, "Target timestamps must be before source timestamps"
    # Interpolate values to find values for each target timestamp without scipy
    return interp1d(source_timestamps, values, axis=0)(target_timestamps)



"""ZOD Frames module."""
import json
import os.path as osp
from itertools import repeat
from pathlib import Path
from typing import Dict, List, Set, Tuple, Union

from tqdm.contrib.concurrent import process_map

from zod.constants import FRAMES, FULL, TRAIN, TRAINVAL_FILES, VAL, VERSIONS
from zod.data_classes.frame import ZodFrame
from zod.data_classes.info import Information
from zod.utils.utils import zfill_id


def _create_frame(frame: dict, dataset_root: str) -> Information:
    info = Information.from_dict(frame)
    info.convert_paths_to_absolute(dataset_root)
    return info


class ZodFrames(object):
    """ZOD Frames."""

    def __init__(self, dataset_root: Union[Path, str], version: str):
        self._dataset_root = dataset_root
        self._version = version
        assert version in VERSIONS, f"Unknown version: {version}, must be one of: {VERSIONS}"
        self._train_frames, self._val_frames = self._load_frames()
        self._frames: Dict[str, Information] = {**self._train_frames, **self._val_frames}

    def __len__(self) -> int:
        return len(self._frames)

    def __getitem__(self, frame_id: Union[int, str, slice]) -> ZodFrame:
        """Get frame by id, which is a 6-digit zero-padded number. Ex: '000001'."""
        if isinstance(frame_id, slice):
            return [self.__getitem__(i) for i in range(*frame_id.indices(len(self)))]
        frame_id = zfill_id(frame_id)
        return ZodFrame(self._frames[frame_id])

    def __iter__(self):
        for frame_id in self._frames:
            yield self.__getitem__(frame_id)

    def _load_frames(self) -> Tuple[Dict[str, Information], Dict[str, Information]]:
        """Load frames for the given version."""
        filename = TRAINVAL_FILES[FRAMES][self._version]

        with open(osp.join(self._dataset_root, filename), "r") as f:
            all_ids = json.load(f)

        train_frames = process_map(
            _create_frame,
            all_ids[TRAIN],
            repeat(self._dataset_root),
            desc="Loading train frames",
            chunksize=50 if self._version == FULL else 1,
        )
        val_frames = process_map(
            _create_frame,
            all_ids[VAL],
            repeat(self._dataset_root),
            desc="Loading val frames",
            chunksize=50 if self._version == FULL else 1,
        )

        train_frames = {frame.id: frame for frame in train_frames}
        val_frames = {frame.id: frame for frame in val_frames}

        return train_frames, val_frames

    def get_split(self, split: str) -> List[str]:
        """Get split by name (e.g. train / val)."""
        if split == TRAIN:
            return list(self._train_frames.keys())
        elif split == VAL:
            return list(self._val_frames.keys())
        else:
            raise ValueError(f"Unknown split: {split}, should be {TRAIN} or {VAL}")

    def get_all_frame_infos(self) -> Dict[str, Information]:
        """Get all frames."""
        return self._frames

    def get_all_ids(self) -> Set[str]:
        """Get all frame ids."""
        return set(self._frames.keys())



"""Functions for evaluating object detection models."""

import json
import os
from itertools import repeat
from typing import Callable, Dict, Iterator, List, Tuple

import numpy as np
from tqdm.contrib.concurrent import process_map

from zod.anno.object import AnnotatedObject, PredictedObject
from zod.constants import EVALUATION_CLASSES
from zod.data_classes.calibration import Calibration
from zod.eval.detection.matching import MatchedFrame, match_one_frame
from zod.eval.detection.nuscenes_eval.common.data_classes import EvalBoxes
from zod.eval.detection.nuscenes_eval.common.utils import center_distance
from zod.eval.detection.nuscenes_eval.detection.algo import accumulate, calc_ap, calc_tp
from zod.eval.detection.nuscenes_eval.detection.data_classes import (
    DetectionBox,
    DetectionConfig,
    DetectionMetrics,
)
from zod.eval.detection.utils import NUSCENES_DEFAULT_SETTINGS

VALID_TP_METRICS = ["trans_err", "scale_err", "orient_err"]
PRECISION_RECALL_SAMPLING_POINTS = 101

NUSCENES_SUMMARY = """
NUSCENES METRIC SUMMARY:
distance function: {dist_func}
distance thresholds: {dist_thresh}
--------------------------------------------------
NDS:  {NDS:.4f}
mAP:  {mAP:.4f}
mATE: {mATE:.4f}
mASE: {mASE:.4f}
mAOE: {mAOE:.4f}"""

PER_CLASS_SUMMARY = """
{cls}:
    mAP:  {mAP:.4f}
    mATE: {mATE:.4f}
    mASE: {mASE:.4f}
    mAOE: {mAOE:.4f}"""

EvalFrame = Tuple[List[AnnotatedObject], List[PredictedObject], Calibration]


class Evalset:
    def __init__(self):
        self._frames: Dict[str, EvalFrame] = {}

    def add_frame(self, frame_id: str, frame: EvalFrame):
        self._frames[frame_id] = frame

    def __len__(self):
        return len(self._frames)

    def __repr__(self) -> str:
        return f"Evalset with {len(self)} frames."

    def __iter__(self) -> Iterator[Tuple[str, EvalFrame]]:
        return iter(self._frames.items())

    def __getitem__(self, frame_id: str) -> EvalFrame:
        return self._frames[frame_id]

    @property
    def frames(self) -> Dict[str, EvalFrame]:
        """Return all frames."""
        return self._frames


def nuscenes_evaluate(
    gt_boxes: EvalBoxes,
    det_boxes: EvalBoxes,
    dist_fcn: Callable = center_distance,
    dist_th: float = 1,
    min_precision: float = 0.1,
    min_recall: float = 0.1,
) -> Dict[str, float]:
    """Perform nuscenes evaluation based on a number of groundtruths and detections."""
    metrics = {}

    for cls in EVALUATION_CLASSES:
        # ensure that we have samples for this class in the gt data
        n_samples = sum(1 for box in gt_boxes.all if box.detection_name == cls)
        if n_samples == 0:
            continue
        md = accumulate(
            gt_boxes,
            det_boxes,
            cls,
            dist_fcn=dist_fcn,
            dist_th=dist_th,
        )

        metrics[cls] = {
            metric: calc_tp(md, min_recall=0.1, metric_name=metric) for metric in VALID_TP_METRICS
        }
        metrics[cls]["ap"] = calc_ap(md, min_recall=min_recall, min_precision=min_precision)
    return metrics


def filter_eval_boxes_on_ranges(
    boxes: EvalBoxes, class_ranges: Dict[str, int], verbose: bool = False
) -> EvalBoxes:
    """Filter out boxes that are outside of the range of the classes."""
    filtered_boxes = EvalBoxes()

    def _filter(box: DetectionBox):
        return (
            box.detection_name in class_ranges
            and class_ranges[box.detection_name][0]
            < box.ego_dist
            <= class_ranges[box.detection_name][1]
        )

    for frame_id in boxes.sample_tokens:
        filtered_boxes.add_boxes(frame_id, [box for box in boxes[frame_id] if _filter(box)])
    if verbose:
        # print the number of removed boxes per class
        for cls in class_ranges:
            n_originally = sum(1 for box in boxes.all if box.detection_name == cls)
            n_filtered = sum(1 for box in filtered_boxes.all if box.detection_name == cls)
            n_removed = n_originally - n_filtered
            if n_removed > 0:
                print(f"{cls}: from {n_originally} -> {n_filtered} ({n_removed} removed)")

    return filtered_boxes


def print_nuscenes_metrics(metrics: DetectionMetrics):
    print(
        NUSCENES_SUMMARY.format(
            dist_func=metrics.cfg.dist_fcn,
            dist_thresh=metrics.cfg.dist_ths,
            NDS=metrics.nd_score,
            mAP=metrics.mean_ap,
            mATE=metrics.tp_errors["trans_err"],
            mASE=metrics.tp_errors["scale_err"],
            mAOE=metrics.tp_errors["orient_err"],
        )
    )
    for cls in metrics.mean_dist_aps:
        print(
            PER_CLASS_SUMMARY.format(
                cls=cls,
                mAP=metrics.mean_dist_aps[cls],
                mATE=metrics._label_tp_errors[cls]["trans_err"],
                mASE=metrics._label_tp_errors[cls]["scale_err"],
                mAOE=metrics._label_tp_errors[cls]["orient_err"],
            )
        )


def original_nuscenes_eval(
    gt_boxes: EvalBoxes,
    det_boxes: EvalBoxes,
    verbose: bool = False,
    output_path: str = None,
) -> DetectionMetrics:
    """Perform nuscenes evaluation based on a number of groundtruths and detections."""
    detection_cfg = DetectionConfig(**NUSCENES_DEFAULT_SETTINGS)
    detection_metrics = DetectionMetrics(detection_cfg)

    class_ranges = {k: (0, v) for k, v in NUSCENES_DEFAULT_SETTINGS["class_range"].items()}

    # filter according to the default nuscenes settings
    gt_boxes = filter_eval_boxes_on_ranges(gt_boxes, class_ranges)
    det_boxes = filter_eval_boxes_on_ranges(det_boxes, class_ranges)

    metrics = {
        dist_th: nuscenes_evaluate(gt_boxes, det_boxes, dist_th=dist_th)
        for dist_th in detection_cfg.dist_ths
    }

    evaluated_clses = set(metrics[detection_cfg.dist_ths[0]].keys())
    for zod_cls in evaluated_clses:
        # They evaluate the ap across all thresholds
        for dist_th in detection_cfg.dist_ths:
            detection_metrics.add_label_ap(
                detection_name=zod_cls,
                dist_th=dist_th,
                ap=metrics[dist_th][zod_cls]["ap"],
            )

        # They evaluate the tp across only one threshold
        for metric in VALID_TP_METRICS:
            detection_metrics.add_label_tp(
                zod_cls, metric, metrics[detection_cfg.dist_th_tp][zod_cls][metric]
            )

    if output_path:
        with open(os.path.join(output_path, "nuscenes_evaluation_metrics.json"), "w") as f:
            json.dump(detection_metrics.serialize(), f)

    if verbose:
        print_nuscenes_metrics(detection_metrics)

    return detection_metrics


def _evaluate_frame(
    frame_id: str,
    frame: EvalFrame,
    cls: str,
    dist_fcn: Callable,
    dist_th: float,
) -> Dict[str, MatchedFrame]:

    gts, preds, calibration = frame
    # filter out all predictions that are not of the correct class
    preds = [pred for pred in preds if pred.name == cls]
    # filter the ground truth boxes based on cls
    gts = [gt for gt in gts if gt.name == cls]

    matched_frame = match_one_frame(
        ground_truth=gts,
        predictions=preds,
        calibration=calibration,
        dist_fcn=dist_fcn,
        dist_threshold=dist_th,
    )

    return {frame_id: matched_frame}


def zod_evaluation(
    evalset: Evalset,
    dist_fcn: Callable,
    dist_th: float,
):
    # evaluate over all classes
    for cls in EVALUATION_CLASSES:
        # process all frames in parallel
        results: List[Dict[str, MatchedFrame]] = process_map(
            _evaluate_frame,
            evalset.frames.keys(),
            evalset.frames.values(),
            repeat(cls),
            repeat(dist_fcn),
            repeat(dist_th),
            desc=f"Accumulating metrics for class: {cls}",
            chunksize=1 if len(evalset.frames) < 100 else 100,
        )

        # join the list of dicts into a single dict
        results: Dict[str, MatchedFrame] = {k: v for d in results for k, v in d.items()}

        # accumulate the metrics across all frames
        tps = []
        fps = []
        confs = []
        n_ground_truths = 0

        for _, matched_frame in results.items():
            n_ground_truths += len(matched_frame.matches) + len(matched_frame.false_negatives)
            for _, pred in matched_frame.matches:
                tps.append(1)
                fps.append(0)
                confs.append(pred.confidence)
            for pred in matched_frame.false_positives:
                tps.append(0)
                fps.append(1)
                confs.append(pred.confidence)

        # sort the accumulated list based on confidence
        confs, tps, fps = zip(*sorted(zip(confs, tps, fps), reverse=True))
        # accumulate the tps and fps
        tps = np.cumsum(tps).astype(np.float)
        fps = np.cumsum(fps).astype(np.float)
        # compute the precision and recall
        precision = tps / (tps + fps)
        recall = tps / n_ground_truths

        raise NotImplementedError("AP computation is not yet implemented")
        # compute the ap
        recall_sampling_points = np.linspace(0, 1, PRECISION_RECALL_SAMPLING_POINTS)
        precision = np.interp(recall_sampling_points, recall, precision, right=0)
        # confidences = np.interp(recall_sampling_points, recall, confs, right=0)
        recall = recall_sampling_points

        # store in some sort of data structure, both the raw data but also the computed metrics
        # todo (william), store the results here
        ap = np.mean(precision)
        print(f"AP for class {cls} is {ap}")

    # compute the final metrics, average over classes etc
    # todo (william), compute the final metrics here



"""Utility functions for object detection evaluation."""

import numpy as np
from shapely import geometry

from zod.anno.object import AnnotatedObject
from zod.data_classes.box import Box2D, Box3D
from zod.eval.detection.nuscenes_eval.detection.data_classes import DetectionBox

NUSCENES_DEFAULT_SETTINGS = {
    "class_range": {
        "Vehicle": 50,
        "VulnerableVehicle": 40,
        "Pedestrian": 30,
        "TrafficSign": 30,
        "TrafficSignal": 30,
    },
    "dist_fcn": "center_distance",
    "dist_ths": [0.5, 1.0, 2.0, 4.0],
    "dist_th_tp": 2.0,
    "min_recall": 0.1,
    "min_precision": 0.1,
    "max_boxes_per_sample": 500,
    "mean_ap_weight": 5,
}


def convert_to_detection_box(frame_id: str, obj: AnnotatedObject) -> DetectionBox:
    """Convert a DynamicObject to a DetectionBox."""
    det_box = DetectionBox(
        sample_token=frame_id,
        translation=tuple(obj.box3d.center),
        size=tuple(obj.box3d.size),
        rotation=tuple(obj.box3d.orientation.elements),
        velocity=(0, 0),  # we dont have velocity in zod
        ego_translation=tuple(0, 0, 0),  # gt is always at (0, 0, 0)
        num_pts=10,
        detection_name=obj.name,
        detection_score=-1.0,  # ground truth boxes have no score
        attribute_name="",  # we dont have attributes in zod
    )

    return det_box


def center_distance(box1: Box3D, box2: Box3D) -> float:
    """Calculate the center distance between two boxes."""
    return np.linalg.norm(box1.center - box2.center)


def center_distance_bev(box1: Box3D, box2: Box3D) -> float:
    """Calculate the center distance between two boxes."""
    assert box1.frame == box2.frame, "Boxes must be in the same frame."
    return np.linalg.norm(box1.center[:2] - box2.center[:2])


def iou3D(box1: Box3D, box2: Box3D) -> float:
    """Calculate the iou between two boxes."""
    assert box1.frame == box2.frame, "Boxes must be in the same frame."

    raise NotImplementedError


def iou2D(box1: Box2D, box2: Box2D) -> float:
    """Calculate the iou between two boxes."""

    # calculate the intersection
    xmin = max(box1.xyxy[0], box2.xyxy[0])
    ymin = max(box1.xyxy[1], box2.xyxy[1])
    xmax = min(box1.xyxy[2], box2.xyxy[2])
    ymax = min(box1.xyxy[3], box2.xyxy[3])
    intersection = max(0, xmax - xmin) * max(0, ymax - ymin)

    # calculate the union
    union = box1.area + box2.area - intersection

    return intersection / union if union > 0.0 else 0.0


def giou2D(box1: Box2D, box2: Box2D) -> float:
    """Calculate the giou between two boxes."""

    # calculate the intersection
    xmin = max(box1.xyxy[0], box2.xyxy[0])
    ymin = max(box1.xyxy[1], box2.xyxy[1])
    xmax = min(box1.xyxy[2], box2.xyxy[2])
    ymax = min(box1.xyxy[3], box2.xyxy[3])
    intersection = max(0, xmax - xmin) * max(0, ymax - ymin)

    # calculate the union
    union = box1.area + box2.area - intersection

    # calculate the convex hull
    xmin = min(box1.xyxy[0], box2.xyxy[0])
    ymin = min(box1.xyxy[1], box2.xyxy[1])
    xmax = max(box1.xyxy[2], box2.xyxy[2])
    ymax = max(box1.xyxy[3], box2.xyxy[3])
    convex_hull = (xmax - xmin) * (ymax - ymin)

    return intersection / union - (convex_hull - union) / convex_hull if union > 0.0 else 0.0


def iod2D(box1: Box2D, detection: Box2D) -> float:
    """Calculate the intersection over detection area between two boxes."""

    # calculate the intersection
    xmin = max(box1.xyxy[0], detection.xyxy[0])
    ymin = max(box1.xyxy[1], detection.xyxy[1])
    xmax = min(box1.xyxy[2], detection.xyxy[2])
    ymax = min(box1.xyxy[3], detection.xyxy[3])
    intersection = max(0, xmax - xmin) * max(0, ymax - ymin)

    # calculate the union
    return intersection / detection.area if detection.area > 0.0 else 0.0


def polygon_iod2D(polygon1: geometry.Polygon, detection: geometry.Polygon) -> float:
    """Calculate the intersection over detection area between two polygons."""
    intersection = polygon1.intersection(detection).area
    return intersection / detection.area if detection.area > 0.0 else 0.0


def polygon_iou2D(polygon1: geometry.Polygon, polygon2: geometry.Polygon) -> float:
    """Calculate the iou between two polygons."""
    intersection = polygon1.intersection(polygon2).area
    union = polygon1.union(polygon2).area
    return intersection / union if union > 0.0 else 0.0


def polygon_giou2D(polygon1: geometry.Polygon, polygon2: geometry.Polygon) -> float:
    """Calculate the giou between two polygons."""
    intersection = polygon1.intersection(polygon2).area
    union = polygon1.union(polygon2).area
    convex_hull = polygon1.convex_hull.union(polygon2.convex_hull).area
    return intersection / union - (convex_hull - union) / convex_hull if union > 0.0 else 0.0



from typing import Optional

import numpy as np

from zod.data_classes.calibration import LidarCalibration
from zod.data_classes.oxts import EgoMotion
from zod.data_classes.sensor import LidarData


def motion_compensate_scanwise(
    lidar_data: LidarData,
    ego_motion: EgoMotion,
    calibration: LidarCalibration,
    target_timestamp: float,
) -> LidarData:
    """Motion compensate a (pointwise compensated) lidar point cloud."""
    lidar_data = lidar_data.copy()
    source_pose = ego_motion.get_poses(lidar_data.core_timestamp)
    target_pose = ego_motion.get_poses(target_timestamp)

    # Compute relative transformation between target pose and source pose
    odometry = np.linalg.inv(target_pose) @ source_pose

    # Project to ego vehicle frame using calib
    lidar_data.transform(calibration.extrinsics)

    # Project to center frame using odometry
    lidar_data.transform(odometry)

    # Project back to lidar frame using calib
    lidar_data.transform(calibration.extrinsics.inverse)
    return lidar_data


def motion_compensate_pointwise(
    lidar_data: LidarData,
    ego_motion: EgoMotion,
    calibration: LidarCalibration,
    target_timestamp: Optional[float] = None,
) -> LidarData:
    """Motion compensate a lidar point cloud in a pointwise manner."""
    lidar_data = lidar_data.copy()
    target_timestamp = target_timestamp or lidar_data.core_timestamp

    # Interpolate oxts data for each frame
    point_poses = ego_motion.get_poses(lidar_data.timestamps)
    target_pose = ego_motion.get_poses(target_timestamp)

    # Compute relative transformation between target pose and point poses
    odometry = np.linalg.inv(target_pose) @ point_poses

    # Project to ego vehicle frame using calib
    lidar_data.transform(calibration.extrinsics)

    # Project to center frame using odometry
    lidar_data.transform(odometry)

    # Project back to lidar frame using calib
    lidar_data.transform(calibration.extrinsics.inverse)
    return lidar_data



"""Top-level package for Zenseact Open Dataset (ZOD)."""

from .constants import AnnotationProject as AnnotationProject
from .constants import Anonymization as Anonymization
from .constants import Camera as Camera
from .constants import Lidar as Lidar
from .zod_frames import ZodFrames as ZodFrames
from .zod_sequences import ZodSequences as ZodSequences

try:
    # importlib.metadata is present in Python 3.8 and later
    import importlib.metadata as importlib_metadata
except ImportError:
    # use the shim package importlib-metadata pre-3.8
    import importlib_metadata as importlib_metadata

__version__ = importlib_metadata.version(__package__ or __name__)



"""Script that prepares the dataset for the traffic sign classification task.

Will create a dataset from the original zod frames by cropping out the relevant traffic signs
in each image and storing them for later use.
"""

import argparse
import json
import os
from dataclasses import dataclass
from itertools import repeat
from typing import Any, Dict, List, Optional, Set, Tuple

import cv2
from tqdm.contrib.concurrent import process_map

import zod.annotations.parser as parser
import zod.constants as constants
from zod import ZodFrames
from zod.data_classes.frame import ZodFrame

SUMMARY = """
Finished creating dataset.
.....................................
Created {} traffic sign samples.
Saved dataset to {}, with a dataset.json file that contains the metadata.

Each image is padded with a factor of {}. Note that the padding might be limited
    by the original image size. The padding is stored in the dataset.json file.

Images are saved in the following folder structure:
    <output_folder>/<traffic_sign_class>/<frame_id>_<annotation_uuid>.png
"""


@dataclass
class Args:
    """Script args."""

    dataset_root: str
    output_folder: str
    num_workers: Optional[int]
    padding_factor: Optional[float]
    padding_px: Optional[Tuple[int, int]]
    overwrite: bool


def _parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument(
        "--output-folder",
        type=str,
        required=True,
        help="Where the dataset should be saved.",
    )
    parser.add_argument(
        "--dataset-root",
        type=str,
        required=True,
        help="Path to the root of the dataset to use.",
    )
    parser.add_argument("--num-workers", type=int, help="Number of workers to use.")
    parser.add_argument(
        "--padding-factor",
        type=float,
        help="Factor to multiply the padding with.",
    )
    parser.add_argument(
        "--padding-px",
        type=int,
        nargs="+",
        help="Padding in x and y direction.",
    )
    parser.add_argument(
        "--overwrite",
        action="store_true",
        help="Whether to overwrite existing files.",
    )

    args = parser.parse_args()

    assert not (
        args.padding_factor is not None and args.padding_px is not None
    ), "Cannot specify both padding and padding_factor"

    return Args(**vars(args))


def _process_frame(frame: ZodFrame, args: Args, train_ids: Set[str]) -> List[Dict[str, Any]]:
    """Process a single frame."""

    traffic_signs: List[parser.TrafficSignAnnotation] = frame.get_annotation(
        constants.AnnotationProject.TRAFFIC_SIGNS
    )

    if len(traffic_signs) == 0:
        return []

    new_cropped_frames = []
    # load the image
    image_path = frame.info.get_key_camera_frame(constants.Anonymization.BLUR).filepath
    image = cv2.cvtColor(cv2.imread(image_path), cv2.COLOR_BGR2RGB)
    for traffic_sign in traffic_signs:

        cls_name = traffic_sign.traffic_sign_class
        train_or_val = constants.TRAIN if frame.frame_id in train_ids else constants.VAL
        cls_folder = os.path.join(args.output_folder, train_or_val, cls_name)

        if not os.path.exists(cls_folder):
            os.makedirs(cls_folder)
            os.chmod(cls_folder, 0o775)

        new_frame_id = f"{frame.frame_id}_{traffic_sign.uuid}"
        output_file = os.path.join(
            cls_folder,
            f"{new_frame_id}.png",
        )
        if not args.overwrite and os.path.exists(output_file):
            continue

        # crop out the correct patch
        cropped_image, padding = traffic_sign.bounding_box.crop_from_image(
            image,
            padding_factor=args.padding_factor,
            padding=args.padding_px,
        )
        # save the image
        cv2.imwrite(output_file, cv2.cvtColor(cropped_image, cv2.COLOR_RGB2BGR))

        original_widht = traffic_sign.bounding_box.dimension[0]
        original_height = traffic_sign.bounding_box.dimension[1]
        center_x = padding[0] + original_widht // 2
        center_y = padding[1] + original_height // 2

        # create a frame info file
        new_cropped_frames.append(
            {
                "frame_id": new_frame_id,
                "image_path": os.path.join(cls_folder, f"{new_frame_id}.png"),
                "padding_left": padding[0],
                "padding_top": padding[1],
                "padding_right": padding[2],
                "padding_bottom": padding[3],
                "padded_width": cropped_image.shape[1],
                "padded_height": cropped_image.shape[0],
                "center_x": center_x,
                "center_y": center_y,
                "original_width": traffic_sign.bounding_box.dimension[0],
                "original_height": traffic_sign.bounding_box.dimension[1],
                "annotation": {
                    key: val for key, val in traffic_sign.__dict__.items() if key != "bounding_box"
                },
            }
        )

    return new_cropped_frames


def main(args: Args):
    """Run the main script."""

    print("Will create dataset from", args.dataset_root)
    print("Will save dataset to", args.output_folder)

    if not os.path.exists((args.output_folder)):
        os.makedirs(args.output_folder)
        os.chmod(args.output_folder, 0o775)
    zod_frames = ZodFrames(args.dataset_root, version="full")
    print(f"Will process {len(zod_frames)} full frames.")

    traffic_sign_frames = process_map(
        _process_frame,
        zod_frames.get_all_frames().values(),
        repeat(args),
        desc="Processing frames in ZOD",
        max_workers=args.num_workers,
        chunksize=1 if args.num_workers == 1 else 100,
    )

    # flatten the returned list
    traffic_sign_frames: List[Dict[str, Any]] = [
        frame for frames in traffic_sign_frames for frame in frames
    ]

    train_frame_ids = set(zod_frames.get_split(constants.TRAIN))
    val_frame_ids = set(zod_frames.get_split(constants.VAL))

    train_frames = [
        f for f in traffic_sign_frames if f["frame_id"].split("_")[0] in train_frame_ids
    ]
    val_frames = [f for f in traffic_sign_frames if f["frame_id"].split("_")[0] in val_frame_ids]

    # write it to a json file
    with open(os.path.join(args.output_folder, "dataset.json"), "w") as f:
        json.dump({constants.TRAIN: train_frames, constants.VAL: val_frames}, f)

    print(
        SUMMARY.format(
            len(traffic_sign_frames),
            args.output_folder,
            args.padding_factor if args.padding_factor is not None else args.padding_px,
        )
    )


if __name__ == "__main__":
    main(_parse_args())
